import xleapp.helpers.strings as strings

from xleapp import Artifact, Search, WebIcon


class GeodPDPlaceCache(Artifact):
    def __post_init__(self):
        self.name = "PD Place Cache"
        self.category = "Geolocation"
        self.web_icon = WebIcon.MAP_PIN
        self.report_headers = (
            "last access time",
            "requestkey",
            "pdplacehash",
            "expire time",
            "pd place",
        )

    @Search("**/com.apple.geod/PDPlaceCache.db")
    def process(self):
        for fp in self.found:
            cursor = fp().cursor()
            cursor.execute(
                """
                SELECT requestkey, pdplacelookup.pdplacehash, datetime('2001-01-01', "lastaccesstime" || ' seconds') as lastaccesstime, datetime('2001-01-01', "expiretime" || ' seconds') as expiretime, pdplace
                FROM pdplacelookup
                INNER JOIN pdplaces on pdplacelookup.pdplacehash = pdplaces.pdplacehash
                """,
            )

            all_rows = cursor.fetchall()
            usageentries = len(all_rows)

            if usageentries > 0:
                data_list = []
                for row in all_rows:
                    pd_place = "".join(f"{row}<br>" for row in set(strings.print(row[4])))
                    data_list.append((row[2], row[0], row[1], row[3], pd_place))
                self.data = data_list



import plistlib

import blackboxprotobuf

from xleapp import Artifact, Search, WebIcon


class AppleMapsApplication(Artifact):
    def __post_init__(self) -> None:
        self.name = "Apple Maps App"
        self.category = "Locations"
        self.web_icon = WebIcon.MAP_PIN
        self.report_headers = ("Latitude", "Longitude")

    @Search("**/Data/Application/*/Library/Preferences/com.apple.Maps.plist")
    def process(self) -> None:
        for fp in self.found:
            pl = plistlib.load(fp())

            types = {
                '1': {'type': 'double', 'name': 'Latitude'},
                '2': {'type': 'double', 'name': 'Longitude'},
                '3': {'type': 'double', 'name': ''},
                '4': {'type': 'fixed64', 'name': ''},
                '5': {'type': 'double', 'name': ''},
            }
            protobuf = pl.get('__internal__LastActivityCamera', None)

            if protobuf:
                internal_plist, di = blackboxprotobuf.decode_message(protobuf, types)
                latitude = internal_plist["Latitude"]
                longitude = internal_plist["Longitude"]

                data_list = []
                data_list.append((latitude, longitude))
        self.data = data_list



# import the protobuf-decoder module
import blackboxprotobuf as pbparser

# import the ccl parser
import xleapp_ios.helpers.parsers.ccl.ccl_bplist as cclparser

# import the ktx parser
import xleapp_ios.helpers.parsers.ktx.ios_ktx2png as ktxparser



from xleapp import Artifact, Search, WebIcon


class AppleWalletTransactions(Artifact):
    def __post_init__(self) -> None:
        self.name = "Tranactions"
        self.category = "Apple Wallet"
        self.web_icon = WebIcon.DOLLAR_SIGN
        self.report_headers = (
            'Transaction Date',
            'Merchant',
            'Locality',
            'Administrative Area',
            'Currency Amount',
            'Currency Type',
            'Location Date',
            'Latitude',
            'Longitude',
            'Altitude',
            'Peer Payment Handle',
            'Payment Memo',
            'Transaction Status',
            'Transaction Type',
        )

    @Search("**/passes23.sqlite")
    def process(self) -> None:
        for fp in self.found:
            cursor = fp().cursor()
            cursor.execute(
                '''SELECT
                        DATETIME(TRANSACTION_DATE + 978307200,'UNIXEPOCH'),
                        MERCHANT_NAME,
                        LOCALITY,
                        ADMINISTRATIVE_AREA,
                        CAST(AMOUNT AS REAL)/100,
                        CURRENCY_CODE,
                        DATETIME(LOCATION_DATE + 978307200,'UNIXEPOCH'),
                        LOCATION_LATITUDE,
                        LOCATION_LONGITUDE,
                        LOCATION_ALTITUDE,
                        PEER_PAYMENT_COUNTERPART_HANDLE,
                        PEER_PAYMENT_MEMO,
                        TRANSACTION_STATUS,
                        TRANSACTION_TYPE
                        FROM PAYMENT_TRANSACTION
                        '''
            )

            all_rows = cursor.fetchall()

        if len(all_rows) > 0:
            data_list = []
            for row in all_rows:
                data_list.append(
                    (
                        row[0],
                        row[1],
                        row[2],
                        row[3],
                        row[4],
                        row[5],
                        row[6],
                        row[7],
                        row[8],
                        row[9],
                        row[10],
                        row[11],
                        row[12],
                        row[13],
                    )
                )



import plistlib

import nska_deserialize as nd

from xleapp import Artifact, Search, WebIcon


class AppItunesMeta(Artifact):
    def __post_init__(self) -> None:
        self.name = "Itunes & Bundle Metadata"
        self.description = "iTunes & Bundle ID Metadata contents for apps"
        self.category = "Installed Apps"
        self.web_icon = WebIcon.PACKAGE
        self.report_headers = (
            "Installed Date",
            "App Purchase Date",
            "Bundle ID",
            "Item Name",
            "Artist Name",
            "Version Number",
            "Downloaded by",
            "Genre",
            "Factory Install",
            "App Release Date",
            "Source App",
            "Sideloaded?",
            "Variant ID",
            "Source File Location",
        )

    @Search("**/iTunesMetadata.plist", "**/BundleMetadata.plist")
    def process(self) -> None:
        metadata, data_list = [], []
        for fp in self.found:
            if fp.path.name == "iTunesMetadata.plist":
                pl = plistlib.load(fp())

                purchasedate = pl.get("com.apple.iTunesStore.downloadInfo", {}).get(
                    "purchaseDate",
                    "",
                )
                bundleid = pl.get("softwareVersionBundleId", "")
                itemname = pl.get("itemName", "")
                artistname = pl.get("artistName", "")
                versionnum = pl.get("bundleShortVersionString", "")
                downloadedby = (
                    pl.get("com.apple.iTunesStore.downloadInfo", {})
                    .get("accountInfo", {})
                    .get("AppleID", "")
                )
                genre = pl.get("genre", "")
                factoryinstall = pl.get("isFactoryInstall", "")
                appreleasedate = pl.get("releaseDate", "")
                sourceapp = pl.get("sourceApp", "")
                sideloaded = pl.get("sideLoadedDeviceBasedVPP", "")
                variantid = pl.get("variantID", "")
                metadata.extend(
                    [
                        purchasedate,
                        bundleid,
                        itemname,
                        artistname,
                        versionnum,
                        downloadedby,
                        genre,
                        factoryinstall,
                        appreleasedate,
                        sourceapp,
                        sideloaded,
                        variantid,
                        fp.path,
                    ],
                )

            if fp.path.name == "BundleMetadata.plist":
                deserialized_plist = nd.deserialize_plist(fp())
                install_date = deserialized_plist.get('installDate', '')

        data_list.append((install_date, *metadata))
        self.data = data_list



from xleapp import Artifact, Search, WebIcon


class ApplePodcastsEpisodes(Artifact):
    def __post_init__(self):
        self.name = "Episodes"
        self.category = "Apple Podcasts"
        self.web_icon = WebIcon.PLAY_CIRCLE
        self.report_headers = (
            'Import Date',
            'Metadata Timestamp',
            'Date Last Played',
            'Play State Last Modified',
            'Download Date',
            'Play Count',
            'Author',
            'Title',
            'Subtitle',
            'Asset URL',
            'Web Page URL',
            'Duration',
            'Size',
            'Play State',
        )

    @Search("**/MTLibrary.sqlite")
    def process(self) -> None:
        for fp in self.found:
            cursor = fp().cursor()
            cursor.execute(
                '''
                SELECT
                datetime(ZIMPORTDATE + 978307200, 'unixepoch'),
                CASE ZMETADATATIMESTAMP
                    WHEN 0 THEN ''
                    ELSE datetime(ZMETADATATIMESTAMP + 978307200, 'unixepoch')
                END,
                datetime(ZLASTDATEPLAYED + 978307200, 'unixepoch'),
                datetime(ZPLAYSTATELASTMODIFIEDDATE + 978307200, 'unixepoch'),
                datetime(ZDOWNLOADDATE + 978307200, 'unixepoch'),
                ZPLAYCOUNT,
                ZAUTHOR,
                ZTITLE,
                ZITUNESSUBTITLE,
                ZASSETURL,
                ZWEBPAGEURL,
                ZDURATION,
                ZBYTESIZE,
                ZPLAYSTATE
                FROM ZMTEPISODE
                ORDER by ZMETADATATIMESTAMP
                '''
            )

        all_rows = cursor.fetchall()
        if len(all_rows) > 0:
            data_list = []
            for row in all_rows:
                data_list.append(
                    (
                        row[0],
                        row[1],
                        row[2],
                        row[3],
                        row[4],
                        row[5],
                        row[6],
                        row[7],
                        row[8],
                        row[9],
                        row[10],
                        row[11],
                        row[12],
                        row[13],
                    ),
                )

        self.data = data_list



import datetime
import plistlib

from xleapp import Artifact, Search, core_artifact


@core_artifact
class LastBuild(Artifact):
    def __post_init__(self):
        self.name = "Last Build"
        self.category = "IOS Build"

    @Search("*LastBuildInfo.plist")
    def process(self):
        data_list = []
        device_info = self.app.device
        for fp in self.found:
            pl = plistlib.load(fp())
            for key, value in pl.items():
                data_list.append((key, value))
                if key in ["ProductVersion", "ProductBuildVersion", "ProductName"]:
                    device_info.update({key: value})
        self.data = data_list


@core_artifact
class ItunesBackupInfo(Artifact):
    def __post_init__(self):
        self.name = "iTunes Backup"
        self.category = "IOS Build"

    @Search("Info.plist")
    def process(self):
        data_list = []
        device_info = self.app.device
        for fp in self.found:
            pl = plistlib.load(fp())
            for key, value in pl.items():
                if (
                    isinstance(value, str)
                    or isinstance(value, int)
                    or isinstance(value, datetime.datetime)
                ):

                    data_list.append((key, value))
                    if key in (
                        "Build Version",
                        "Device Name",
                        "ICCID",
                        "IMEI",
                        "Last Backup Date",
                        "MEID",
                        "Phone Number",
                        "Product Name",
                        "Product Type",
                        "Product Version",
                        "Serial Number",
                    ):
                        device_info.update({key: value})

                elif key == "Installed Applications":
                    data_list.append((key, ", ".join(value)))
        self.data = data_list



"""
Copyright (c) 2012-2016, CCL Forensics
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the CCL Forensics nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL CCL FORENSICS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import sys
import os
import struct
import datetime

__version__ = "0.21"
__description__ = (
    "Converts Apple binary PList files into a native Python data structure"
)
__contact__ = "Alex Caithness"

_object_converter = None


def set_object_converter(function):
    """Sets the object converter function to be used when retrieving objects from the bplist.
    default is None (which will return objects in their raw form).
    A built in converter (ccl_bplist.NSKeyedArchiver_common_objects_convertor) which is geared
    toward dealling with common types in NSKeyedArchiver is available which can simplify code greatly
    when dealling with these types of files."""
    if not hasattr(function, "__call__"):
        raise TypeError("function is not a function")
    global _object_converter
    _object_converter = function


class BplistError(Exception):
    pass


class BplistUID:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "UID: {0}".format(self.value)

    def __str__(self):
        return self.__repr__()


def __decode_multibyte_int(b, signed=True):
    if len(b) == 1:
        fmt = ">B"  # Always unsigned?
    elif len(b) == 2:
        fmt = ">h"
    elif len(b) == 3:
        if signed:
            return ((b[0] << 16) | struct.unpack(">H", b[1:])[0]) - (
                (b[0] >> 7) * 2 * 0x800000
            )
        else:
            return (b[0] << 16) | struct.unpack(">H", b[1:])[0]
    elif len(b) == 4:
        fmt = ">i"
    elif len(b) == 8:
        fmt = ">q"
    elif len(b) == 16:
        # special case for BigIntegers
        high, low = struct.unpack(">QQ", b)
        result = (high << 64) | low
        if high & 0x8000000000000000 and signed:
            result -= 0x100000000000000000000000000000000
        return result
    else:
        raise BplistError("Cannot decode multibyte int of length {0}".format(len(b)))

    if signed and len(b) > 1:
        return struct.unpack(fmt.lower(), b)[0]
    else:
        return struct.unpack(fmt.upper(), b)[0]


def __decode_float(b, signed=True):
    if len(b) == 4:
        fmt = ">f"
    elif len(b) == 8:
        fmt = ">d"
    else:
        raise BplistError("Cannot decode float of length {0}".format(len(b)))

    if signed:
        return struct.unpack(fmt.lower(), b)[0]
    else:
        return struct.unpack(fmt.upper(), b)[0]


def __decode_object(f, offset, collection_offset_size, offset_table):
    # Move to offset and read type
    # print("Decoding object at offset {0}".format(offset))
    f.seek(offset)
    # A little hack to keep the script portable between py2.x and py3k
    if sys.version_info[0] < 3:
        type_byte = ord(f.read(1)[0])
    else:
        type_byte = f.read(1)[0]
    # print("Type byte: {0}".format(hex(type_byte)))
    if type_byte == 0x00:  # Null      0000 0000
        return None
    elif type_byte == 0x08:  # False   0000 1000
        return False
    elif type_byte == 0x09:  # True    0000 1001
        return True
    elif type_byte == 0x0F:  # Fill    0000 1111
        raise BplistError(
            "Fill type not currently supported at offset {0}".format(f.tell())
        )  # Not sure what to return really...
    elif type_byte & 0xF0 == 0x10:  # Int    0001 xxxx
        int_length = 2 ** (type_byte & 0x0F)
        int_bytes = f.read(int_length)
        return __decode_multibyte_int(int_bytes)
    elif type_byte & 0xF0 == 0x20:  # Float   0010 nnnn
        float_length = 2 ** (type_byte & 0x0F)
        float_bytes = f.read(float_length)
        return __decode_float(float_bytes)
    elif type_byte & 0xFF == 0x33:  # Date   0011 0011
        date_bytes = f.read(8)
        date_value = __decode_float(date_bytes)
        try:
            result = datetime.datetime(2001, 1, 1) + datetime.timedelta(
                seconds=date_value
            )
        except OverflowError:
            result = datetime.datetime.min
        return result
    elif type_byte & 0xF0 == 0x40:  # Data   0100 nnnn
        if type_byte & 0x0F != 0x0F:
            # length in 4 lsb
            data_length = type_byte & 0x0F
        else:
            # A little hack to keep the script portable between py2.x and py3k
            if sys.version_info[0] < 3:
                int_type_byte = ord(f.read(1)[0])
            else:
                int_type_byte = f.read(1)[0]
            if int_type_byte & 0xF0 != 0x10:
                raise BplistError(
                    "Long Data field definition not followed by int type at offset {0}".format(
                        f.tell()
                    )
                )
            int_length = 2 ** (int_type_byte & 0x0F)
            int_bytes = f.read(int_length)
            data_length = __decode_multibyte_int(int_bytes, False)
        return f.read(data_length)
    elif type_byte & 0xF0 == 0x50:  # ASCII  0101 nnnn
        if type_byte & 0x0F != 0x0F:
            # length in 4 lsb
            ascii_length = type_byte & 0x0F
        else:
            # A little hack to keep the script portable between py2.x and py3k
            if sys.version_info[0] < 3:
                int_type_byte = ord(f.read(1)[0])
            else:
                int_type_byte = f.read(1)[0]
            if int_type_byte & 0xF0 != 0x10:
                raise BplistError(
                    "Long ASCII field definition not followed by int type at offset {0}".format(
                        f.tell()
                    )
                )
            int_length = 2 ** (int_type_byte & 0x0F)
            int_bytes = f.read(int_length)
            ascii_length = __decode_multibyte_int(int_bytes, False)
        return f.read(ascii_length).decode("ascii")
    elif type_byte & 0xF0 == 0x60:  # UTF-16  0110 nnnn
        if type_byte & 0x0F != 0x0F:
            # length in 4 lsb
            utf16_length = (type_byte & 0x0F) * 2  # Length is characters - 16bit width
        else:
            # A little hack to keep the script portable between py2.x and py3k
            if sys.version_info[0] < 3:
                int_type_byte = ord(f.read(1)[0])
            else:
                int_type_byte = f.read(1)[0]
            if int_type_byte & 0xF0 != 0x10:
                raise BplistError(
                    "Long UTF-16 field definition not followed by int type at offset {0}".format(
                        f.tell()
                    )
                )
            int_length = 2 ** (int_type_byte & 0x0F)
            int_bytes = f.read(int_length)
            utf16_length = __decode_multibyte_int(int_bytes, False) * 2
        return f.read(utf16_length).decode("utf_16_be")
    elif type_byte & 0xF0 == 0x80:  # UID    1000 nnnn
        uid_length = (type_byte & 0x0F) + 1
        uid_bytes = f.read(uid_length)
        return BplistUID(__decode_multibyte_int(uid_bytes, signed=False))
    elif type_byte & 0xF0 == 0xA0:  # Array  1010 nnnn
        if type_byte & 0x0F != 0x0F:
            # length in 4 lsb
            array_count = type_byte & 0x0F
        else:
            # A little hack to keep the script portable between py2.x and py3k
            if sys.version_info[0] < 3:
                int_type_byte = ord(f.read(1)[0])
            else:
                int_type_byte = f.read(1)[0]
            if int_type_byte & 0xF0 != 0x10:
                raise BplistError(
                    "Long Array field definition not followed by int type at offset {0}".format(
                        f.tell()
                    )
                )
            int_length = 2 ** (int_type_byte & 0x0F)
            int_bytes = f.read(int_length)
            array_count = __decode_multibyte_int(int_bytes, signed=False)
        array_refs = []
        for i in range(array_count):
            array_refs.append(
                __decode_multibyte_int(f.read(collection_offset_size), False)
            )
        return [
            __decode_object(
                f, offset_table[obj_ref], collection_offset_size, offset_table
            )
            for obj_ref in array_refs
        ]
    elif type_byte & 0xF0 == 0xC0:  # Set  1010 nnnn
        if type_byte & 0x0F != 0x0F:
            # length in 4 lsb
            set_count = type_byte & 0x0F
        else:
            # A little hack to keep the script portable between py2.x and py3k
            if sys.version_info[0] < 3:
                int_type_byte = ord(f.read(1)[0])
            else:
                int_type_byte = f.read(1)[0]
            if int_type_byte & 0xF0 != 0x10:
                raise BplistError(
                    "Long Set field definition not followed by int type at offset {0}".format(
                        f.tell()
                    )
                )
            int_length = 2 ** (int_type_byte & 0x0F)
            int_bytes = f.read(int_length)
            set_count = __decode_multibyte_int(int_bytes, signed=False)
        set_refs = []
        for i in range(set_count):
            set_refs.append(
                __decode_multibyte_int(f.read(collection_offset_size), False)
            )
        return [
            __decode_object(
                f, offset_table[obj_ref], collection_offset_size, offset_table
            )
            for obj_ref in set_refs
        ]
    elif type_byte & 0xF0 == 0xD0:  # Dict  1011 nnnn
        if type_byte & 0x0F != 0x0F:
            # length in 4 lsb
            dict_count = type_byte & 0x0F
        else:
            # A little hack to keep the script portable between py2.x and py3k
            if sys.version_info[0] < 3:
                int_type_byte = ord(f.read(1)[0])
            else:
                int_type_byte = f.read(1)[0]
            # print("Dictionary length int byte: {0}".format(hex(int_type_byte)))
            if int_type_byte & 0xF0 != 0x10:
                raise BplistError(
                    "Long Dict field definition not followed by int type at offset {0}".format(
                        f.tell()
                    )
                )
            int_length = 2 ** (int_type_byte & 0x0F)
            int_bytes = f.read(int_length)
            dict_count = __decode_multibyte_int(int_bytes, signed=False)
        key_refs = []
        # print("Dictionary count: {0}".format(dict_count))
        for i in range(dict_count):
            key_refs.append(
                __decode_multibyte_int(f.read(collection_offset_size), False)
            )
        value_refs = []
        for i in range(dict_count):
            value_refs.append(
                __decode_multibyte_int(f.read(collection_offset_size), False)
            )

        dict_result = {}
        for i in range(dict_count):
            # print("Key ref: {0}\tVal ref: {1}".format(key_refs[i], value_refs[i]))
            key = __decode_object(
                f, offset_table[key_refs[i]], collection_offset_size, offset_table
            )
            val = __decode_object(
                f, offset_table[value_refs[i]], collection_offset_size, offset_table
            )
            dict_result[key] = val
        return dict_result


def load(f):
    """
    Reads and converts a file-like object containing a binary property list.
    Takes a file-like object (must support reading and seeking) as an argument
    Returns a data structure representing the data in the property list
    """
    # Check magic number
    if f.read(8) != b"bplist00":
        raise BplistError("Bad file header")

    # Read trailer
    f.seek(-32, os.SEEK_END)
    trailer = f.read(32)
    (
        offset_int_size,
        collection_offset_size,
        object_count,
        top_level_object_index,
        offest_table_offset,
    ) = struct.unpack(">6xbbQQQ", trailer)

    # Read offset table
    f.seek(offest_table_offset)
    offset_table = []
    for i in range(object_count):
        offset_table.append(__decode_multibyte_int(f.read(offset_int_size), False))

    return __decode_object(
        f, offset_table[top_level_object_index], collection_offset_size, offset_table
    )


def NSKeyedArchiver_common_objects_convertor(o):
    """Built in converter function (suitable for submission to set_object_converter()) which automatically
    converts the following common data-types found in NSKeyedArchiver:
    NSDictionary/NSMutableDictionary;
    NSArray/NSMutableArray;
    NSSet/NSMutableSet
    NSString/NSMutableString
    NSDate
    $null strings"""
    # Conversion: NSDictionary
    if is_nsmutabledictionary(o):
        return convert_NSMutableDictionary(o)
    # Conversion: NSArray
    elif is_nsarray(o):
        return convert_NSArray(o)
    elif is_isnsset(o):
        return convert_NSSet(o)
    # Conversion: NSString
    elif is_nsstring(o):
        return convert_NSString(o)
    # Conversion: NSDate
    elif is_nsdate(o):
        return convert_NSDate(o)
    # Conversion: "$null" string
    elif isinstance(o, str) and o == "$null":
        return None
    # Fallback:
    else:
        return o


def NSKeyedArchiver_convert(o, object_table):
    if isinstance(o, list):
        # return NsKeyedArchiverList(o, object_table)
        result = NsKeyedArchiverList(o, object_table)
    elif isinstance(o, dict):
        # return NsKeyedArchiverDictionary(o, object_table)
        result = NsKeyedArchiverDictionary(o, object_table)
    elif isinstance(o, BplistUID):
        # return NSKeyedArchiver_convert(object_table[o.value], object_table)
        result = NSKeyedArchiver_convert(object_table[o.value], object_table)
    else:
        # return o
        result = o

    if _object_converter:
        return _object_converter(result)
    else:
        return result


class NsKeyedArchiverDictionary(dict):
    def __init__(self, original_dict, object_table):
        super(NsKeyedArchiverDictionary, self).__init__(original_dict)
        self.object_table = object_table

    def __getitem__(self, index):
        o = super(NsKeyedArchiverDictionary, self).__getitem__(index)
        return NSKeyedArchiver_convert(o, self.object_table)

    def get(self, key, default=None):
        return self[key] if key in self else default

    def values(self):
        for k in self:
            yield self[k]

    def items(self):
        for k in self:
            yield k, self[k]


class NsKeyedArchiverList(list):
    def __init__(self, original_iterable, object_table):
        super(NsKeyedArchiverList, self).__init__(original_iterable)
        self.object_table = object_table

    def __getitem__(self, index):
        o = super(NsKeyedArchiverList, self).__getitem__(index)
        return NSKeyedArchiver_convert(o, self.object_table)

    def __iter__(self):
        for o in super(NsKeyedArchiverList, self).__iter__():
            yield NSKeyedArchiver_convert(o, self.object_table)


def deserialise_NsKeyedArchiver(obj, parse_whole_structure=False):
    """Deserialises an NSKeyedArchiver bplist rebuilding the structure.
    obj should usually be the top-level object returned by the load()
    function."""

    # Check that this is an archiver and version we understand
    if not isinstance(obj, dict):
        raise TypeError("obj must be a dict")
    if "$archiver" not in obj or obj["$archiver"] not in (
        "NSKeyedArchiver",
        "NRKeyedArchiver",
    ):
        raise ValueError(
            "obj does not contain an '$archiver' key or the '$archiver' is unrecognised"
        )
    if "$version" not in obj or obj["$version"] != 100000:
        raise ValueError(
            "obj does not contain a '$version' key or the '$version' is unrecognised"
        )

    object_table = obj["$objects"]
    if "root" in obj["$top"] and not parse_whole_structure:
        return NSKeyedArchiver_convert(obj["$top"]["root"], object_table)
    else:
        return NSKeyedArchiver_convert(obj["$top"], object_table)


# NSMutableDictionary convenience functions
def is_nsmutabledictionary(obj):
    if not isinstance(obj, dict):
        return False
    if "$class" not in obj.keys():
        return False
    if obj["$class"].get("$classname") not in ("NSMutableDictionary", "NSDictionary"):
        return False
    if "NS.keys" not in obj.keys():
        return False
    if "NS.objects" not in obj.keys():
        return False

    return True


def convert_NSMutableDictionary(obj):
    """Converts a NSKeyedArchiver serialised NSMutableDictionary into
    a straight dictionary (rather than two lists as it is serialised
    as)"""

    # The dictionary is serialised as two lists (one for keys and one
    # for values) which obviously removes all convenience afforded by
    # dictionaries. This function converts this structure to an
    # actual dictionary so that values can be accessed by key.

    if not is_nsmutabledictionary(obj):
        raise ValueError(
            "obj does not have the correct structure for a NSDictionary/NSMutableDictionary serialised to a NSKeyedArchiver"
        )
    keys = obj["NS.keys"]
    vals = obj["NS.objects"]

    # sense check the keys and values:
    if not isinstance(keys, list):
        raise TypeError(
            "The 'NS.keys' value is an unexpected type (expected list; actual: {0}".format(
                type(keys)
            )
        )
    if not isinstance(vals, list):
        raise TypeError(
            "The 'NS.objects' value is an unexpected type (expected list; actual: {0}".format(
                type(vals)
            )
        )
    if len(keys) != len(vals):
        raise ValueError(
            "The length of the 'NS.keys' list ({0}) is not equal to that of the 'NS.objects ({1})".format(
                len(keys), len(vals)
            )
        )

    result = {}
    for i, k in enumerate(keys):
        if k in result:
            raise ValueError("The 'NS.keys' list contains duplicate entries")
        result[k] = vals[i]

    return result


# NSArray convenience functions
def is_nsarray(obj):
    if not isinstance(obj, dict):
        return False
    if "$class" not in obj.keys():
        return False
    if obj["$class"].get("$classname") not in ("NSArray", "NSMutableArray"):
        return False
    if "NS.objects" not in obj.keys():
        return False

    return True


def convert_NSArray(obj):
    if not is_nsarray(obj):
        raise ValueError(
            "obj does not have the correct structure for a NSArray/NSMutableArray serialised to a NSKeyedArchiver"
        )

    return obj["NS.objects"]


# NSSet convenience functions
def is_isnsset(obj):
    if not isinstance(obj, dict):
        return False
    if "$class" not in obj.keys():
        return False
    if obj["$class"].get("$classname") not in ("NSSet", "NSMutableSet"):
        return False
    if "NS.objects" not in obj.keys():
        return False

    return True


def convert_NSSet(obj):
    if not is_isnsset(obj):
        raise ValueError(
            "obj does not have the correct structure for a NSSet/NSMutableSet serialised to a NSKeyedArchiver"
        )

    return list(obj["NS.objects"])


# NSString convenience functions
def is_nsstring(obj):
    if not isinstance(obj, dict):
        return False
    if "$class" not in obj.keys():
        return False
    if obj["$class"].get("$classname") not in ("NSString", "NSMutableString"):
        return False
    if "NS.string" not in obj.keys():
        return False
    return True


def convert_NSString(obj):
    if not is_nsstring(obj):
        raise ValueError(
            "obj does not have the correct structure for a NSString/NSMutableString serialised to a NSKeyedArchiver"
        )

    return obj["NS.string"]


# NSDate convenience functions
def is_nsdate(obj):
    if not isinstance(obj, dict):
        return False
    if "$class" not in obj.keys():
        return False
    if obj["$class"].get("$classname") not in ("NSDate"):
        return False
    if "NS.time" not in obj.keys():
        return False

    return True


def convert_NSDate(obj):
    if not is_nsdate(obj):
        raise ValueError(
            "obj does not have the correct structure for a NSDate serialised to a NSKeyedArchiver"
        )

    return datetime.datetime(2001, 1, 1) + datetime.timedelta(seconds=obj["NS.time"])



"""
    (c) Yogesh Khatri 2020, MIT License

    This library converts ios created KTX files to PNG. These are 
    textures in KTX format, with raw texture  data in compressed ATSC
    format using a block size of 4x4, which is further compressed with
    LZFSE compression.

    iOS created KTX files use Adaptive Scalable Texture Compression 
    (ASTC) as the compressed texture format. Note that a KTX file
    can contain other formats too like PVR and several other 
    variations of ASTC which can't be read by this code. Here we
    only focus on ASTC 4x4, which is the default on iOS. Some apps 
    ship with KTX files that are not of the same type and can't be
    parsed with this code.

    Dependencies
    ------------
    pillow, pyliblzfse, astc_decomp

    Install deps via pip:
      pip3 install pyliblzfse astc_decomp pillow

    Usage
    -----

    python3 ios_ktx2png.py SAMPLE.KTX

    Output will be in the same folder, called SAMPLE.KTX.png

    See main 
"""

import astc_decomp
import liblzfse
import os
import struct
import sys

from PIL import Image

version = 1.0


class KTX_reader:
    def __init__(self):
        # KTX header fields
        self.identifier = b""  # b"«KTX 11»\r\n\x1A\x0A"
        self.endianness = ""
        self.glType = 0
        self.glTypeSize = 0
        self.glFormat = 0
        self.glInternalFormat = 0
        self.glBaseInternalFormat = 0
        self.pixelWidth = 0
        self.pixelHeight = 0
        self.pixelDepth = 0
        self.numberOfArrayElements = 0
        self.numberOfFaces = 0
        self.numberOfMipmapLevels = 0
        self.bytesOfKeyValueData = 0

        self.error_message = ""
        self.is_aapl_file = False
        self.aapl_data_pos = 0
        self.aapl_data_size = 0
        self.aapl_is_compressed = False

    def validate_header(self, f):
        """Reads header and validates version

        Arguments
        ---------
        f : file-like object

        Returns
        -------
        Bool : True if file can be parsed else False
        """
        f.seek(0)
        header = f.read(0x40)
        if len(header) < 0x40:
            self.error_message = "File too small or can't read"
            return False

        self.identifier = header[0:12]
        self.endianness = header[12:16]
        if self.endianness == bytes.fromhex("01020304"):
            endianness = "<"
        else:
            endianness = ">"
        if self.identifier[0:7] == b"\xabKTX 11":  # good
            (
                self.glType,
                self.glTypeSize,
                self.glFormat,
                self.glInternalFormat,
                self.glBaseInternalFormat,
                self.pixelWidth,
                self.pixelHeight,
                self.pixelDepth,
                self.numberOfArrayElements,
                self.numberOfFaces,
                self.numberOfMipmapLevels,
                self.bytesOfKeyValueData,
            ) = struct.unpack(endianness + "12I", header[16:64])
            return True
        elif self.identifier[0:4] == b"\xabKTX":  # different version
            self.error_message = "Unknown KTX version"
        elif self.identifier[0:8] == b"AAPL\x0D\x0A\x1A\x0A":  # Not KTX, but similar!!
            self.is_aapl_file = True
            return self.parse_aapl_file(f)
        else:
            self.error_message = "Not a KTX file"
        return False

    def parse_aapl_file(self, f):
        """Returns True if HEAD was found and header items could be read"""
        ret = False
        next_header_pos = 8
        f.seek(next_header_pos)
        data = f.read(8)
        while data:
            item_size = struct.unpack("<I", data[0:4])[0]
            item_identifier = data[4:8]
            if item_identifier == b"HEAD":
                item_data = f.read(item_size)
                # read metadata here...
                (
                    _,
                    _,
                    _,
                    _,
                    self.glInternalFormat,
                    self.glBaseInternalFormat,
                    self.pixelWidth,
                    self.pixelHeight,
                    self.pixelDepth,
                    self.numberOfArrayElements,
                    self.numberOfFaces,
                ) = struct.unpack("<11I", item_data[0:44])
                ret = True
            elif item_identifier == b"LZFS":
                self.aapl_data_pos = f.tell() + 4
                self.aapl_data_size = item_size - 4
                self.aapl_is_compressed = True
            elif item_identifier == b"astc":
                self.aapl_data_pos = f.tell() + 4
                self.aapl_data_size = item_size - 4
            next_header_pos += 8 + item_size
            f.seek(next_header_pos)
            data = f.read(8)
        return ret

    def get_uncompressed_texture_data(self, f):
        """Just read the texture data which is lzfse compressed, uncompress and return it.
        Exceptions raised are ValueError or liblzfse.error
        """
        if self.glInternalFormat == 0x93B0:
            if self.is_aapl_file:
                f.seek(self.aapl_data_pos)
                data = f.read(self.aapl_data_size)
                if self.aapl_is_compressed:
                    decompressed = liblzfse.decompress(data)
                    return decompressed
                else:
                    return data
            else:
                f.seek(0x40)
                k_v_data = f.read(self.bytesOfKeyValueData)
                compressed = True if k_v_data.find(b"Compression_APPLE") >= 0 else False
                f.seek(0x40 + self.bytesOfKeyValueData)
                data = f.read()
                if compressed:
                    if data[12:15] == b"bvx":
                        decompressed = liblzfse.decompress(data[12:])
                        return decompressed
                    else:
                        raise ValueError("Unsupported compression, not lzfse!")
                else:
                    return data[
                        4:
                    ]  # first 4 bytes is size (which is practically rest of file)
        else:
            raise ValueError("Unsupported Format")
        return b""

    def convert_to_png(self, f, save_to_path):
        # using astc_decomp
        """Exports KTX as a PNG file
        Arguments
        ---------
        f            : file-like object

        save_to_path : Path of file to save to. If file exists, it will be
        overwritten.

        Returns
        -------
        Bool : True if file was successfully exported else False

        Raises
        ----------
        OSError, ValueError, liblzfse.error
        """
        if self.validate_header(f):
            data = self.get_uncompressed_texture_data(f)
            dec_img = Image.frombytes(
                "RGBA", (self.pixelWidth, self.pixelHeight), data, "astc", (4, 4, False)
            )
            dec_img.save(save_to_path, "PNG")
            return True
        return False

    def save_uncompressed_texture(self, f, save_to_path):
        if self.validate_header(f):
            data = self.get_uncompressed_texture_data(f)
            f = open(save_to_path, "wb")
            f.write(data)
            f.close()
            return True
        return False


def main():
    if sys.argv[0].lower().endswith(".exe"):
        executor = os.path.basename(sys.argv[0])
    else:
        executor = "python3 " + os.path.basename(sys.argv[0])
    usage = (
        "\nios_ktx2png ver {0} - converts ios created KTX to PNG\n"
        + "  (c) 2020 Yogesh Khatri MIT License\n\n"
        + "Usage:\n-----\n  {1} SAMPLE.KTX\n\n"
        + "Output will be stored as SAMPLE.KTX.png in same folder\n"
    )
    usage = usage.format(version, executor)
    argc = len(sys.argv)
    if argc < 2:
        print("Exiting.. Not enough arguments. See usage below.")
        print(usage)
        return
    elif ("-h" in sys.argv) or ("--help" in sys.argv):
        print(usage)
        return
    else:
        ktx_path = sys.argv[1]
        with open(ktx_path, "rb") as f:
            ktx = KTX_reader()
            # ktx.save_uncompressed_texture(f, ktx_path + '.astc') # If you wish to inspect raw data
            try:
                if ktx.convert_to_png(f, ktx_path + ".png"):
                    print(f"Conversion successful!\nOutput is at {ktx_path}.png")
                else:
                    print(
                        "Conversion to PNG failed "
                        + (ktx.error_message if ktx.error_message else "")
                    )
            except (OSError, ValueError, liblzfse.error) as ex:
                print(f"Had an exception - {str(ex)}")


if __name__ == "__main__":
    main()



import datetime

from pathlib import Path

from PIL import Image
from xleapp import Artifact, Search, WebIcon

from xleapp_ios.helpers.parsers import ktxparser


class ApplicationSnapshots(Artifact):
    def __post_init__(self):
        self.name = "App Snapshots (screenshots)"
        self.category = "Installed Apps"
        self.web_icon = WebIcon.PACKAGE
        self.description = (
            "Snapshots saved by iOS for individual apps appear here. Blank "
            "screenshots are excluded here. Dates and times shown are from "
            "file modified timestamps"
        )
        self.report_headers = ('App Name', 'Source Path', 'Date Modified', 'Snapshot')

    @Search(
        "**/Library/Caches/Snapshots/*",
        "**/SplashBoard/Snapshots/*",
        file_names_only=True,
        return_on_first_hit=False,
    )
    def process(self) -> None:
        def save_ktx_to_png_if_valid(ktx_path: Path, save_to_path: Path) -> bool:
            """Excludes all white or all black blank images"""
            with open(ktx_path, 'rb') as f:
                ktx = ktxparser.KTX_reader()
                try:
                    if ktx.validate_header(f):
                        data = ktx.get_uncompressed_texture_data(f)
                        dec_img = Image.frombytes(
                            'RGBA',
                            (ktx.pixelWidth, ktx.pixelHeight),
                            data,
                            'astc',
                            (4, 4, False),
                        )
                        # either all black or all white https://stackoverflow.com/questions/14041562/python-pil-detect-if-an-image-is-completely-black-or-white
                        # if sum(dec_img.convert("L").getextrema()) in (0, 2):
                        #     logfunc('Skipping image as it is blank')
                        #     return False

                        dec_img.save(save_to_path, "PNG")
                        return True
                except (OSError, ValueError, ktxparser.liblzfse.error) as ex:
                    self.log(message=f'Had an exception - {str(ex)}')
            return False

        data_list = []
        # Format=  [ [ 'App Name', 'ktx_path', mod_date, 'png_path' ], .. ]

        for fp in self.found:
            if fp.path.is_dir():
                continue

            if fp.path.stat().st_size < 2500:
                continue

            if "downscaled" not in fp.path.parts:
                app_name = fp.path.parts[-2].split(" ")[0]
            else:
                app_name = fp.path.parts[-3].split(" ")[0]

            if app_name.startswith('sceneID'):
                app_name = app_name[8:]

            dash_pos = app_name.find("-")

            if dash_pos > 0:
                app_name = app_name[0:dash_pos]

            if fp.path.suffix == '.ktx':
                if fp.path.stat().st_size < 2500:  # too small, they are blank
                    continue
                png_path = Path(
                    self.data_save_folder / f"{app_name}_{fp.path.parts[-1][:4]}.png",
                )
                if save_ktx_to_png_if_valid(fp.path, png_path):
                    last_modified_date = datetime.datetime.fromtimestamp(
                        fp.path.stat().st_mtime,
                    )
                    data_list.append([app_name, fp.path, last_modified_date, png_path])
            elif fp.path.suffix == '.jpeg':
                jpg_path = Path(self.data_save_folder / f"{app_name}_{fp.path.parts[-1]}")
                if self.copyfile(fp.path, jpg_path):
                    last_modified_date = datetime.datetime.fromtimestamp(
                        fp.path.stat().st_mtime,
                    )
                    data_list.append([app_name, fp.path, last_modified_date, jpg_path])

        if len(data_list) > 0:
            data_list_for_report = []
            for app_name, ktx_path, mod_date, img_path in data_list:
                img_html = (
                    f'<a href="{str(img_path)}"><img src="{str(img_path)}" '
                    'class="img-fluid" style="max-height:300px; max-width:400px"></a>'
                )
                data_list_for_report.append(
                    (f"{str(app_name)}", f"{str(ktx_path)}", mod_date, img_html),
                )
            self.data = data_list_for_report



import plistlib

from xleapp import Artifact, Search, WebIcon


class AccountConfiguration(Artifact):
    def __post_init__(self):
        self.name = "Account Configuration"
        self.category = "Accounts"
        self.web_icon = WebIcon.USER

    @Search("**/com.apple.accounts.exists.plist")
    def process(self):
        data_list = []
        for fp in self.found:
            pl = plistlib.load(fp())
            for key, value in pl.items():
                data_list.append((key, value))

        self.data = data_list



import logging
import plistlib

from xleapp import Artifact, Search, WebIcon

from ..helpers.utils import bytes_to_mac_address


class AppleWifiScannedPrivateNetworks(Artifact):
    def __post_init__(self):
        self.name = "Wifi Networks Scanned (private)"
        self.description = 'WiFi networks scanned while using fake ("private") MAC address. Dates are taken straight from the source plist.'
        self.category = "Locations"
        self.report_headers = (
            "SSID",
            "BSSID",
            "Added At",
            "Last Joined",
            "Last Updated",
            "MAC Used For Network",
            "Private MAC Computed For Network",
            "MAC Valid",
            "In Known Networks",
            "File",
        )
        self.report_title = "Wifi Networks Scanned (private)"
        self.web_icon = WebIcon.WIFI

    @Search(
        "**/com.apple.wifi.plist",
        "**/com.apple.wifi-networks.plist.backup",
        "**/com.apple.wifi.known-networks.plist",
        "**/com.apple.wifi-private-mac-networks.plist",
    )
    def process(self):
        class ScannedNetwork:
            ssid: str
            bssid: str
            added_at: str
            last_joined: str
            last_updated: str
            private_mac_in_use: str = ""
            private_mac_value: str = ""
            private_mac_valid: str = ""
            in_known_networks: str
            last_auto_joined: str
            plist: str

            def __init__(self, network: dict) -> None:
                self.ssid = str(network.get("SSID_STR", ""))
                self.bssid = str(network.get("BSSID", ""))
                self.last_updated = str(network.get("lastUpdated", ""))
                self.last_joined = str(network.get("lastJoined", ""))
                self.added_at = str(network.get("addedAt", ""))
                self.in_known_networks = str(
                    network.get(
                        "PresentInKnownNetworks",
                        "",
                    ),
                )

                private_mac_address = network.get("PRIVATE_MAC_ADDRESS", "")
                if private_mac_address:
                    self.private_mac_in_use = str(
                        bytes_to_mac_address(
                            private_mac_address.get("PRIVATE_MAC_ADDRESS_IN_USE", ""),
                        ),
                    )
                    self.private_mac_value = (
                        str(
                            bytes_to_mac_address(
                                private_mac_address.get("PRIVATE_MAC_ADDRESS_VALUE", ""),
                            ),
                        ),
                    )
                    self.private_mac_valid = (
                        str(
                            private_mac_address.get("PRIVATE_MAC_ADDRESS_VALID", ""),
                        ),
                    )

            def attributes(self) -> list[str]:
                return [
                    self.ssid,
                    self.bssid,
                    self.added_at,
                    self.last_joined,
                    self.last_updated,
                    self.private_mac_in_use,
                    self.private_mac_value,
                    self.private_mac_valid,
                    self.in_known_networks,
                    self.last_auto_joined,
                    self.plist,
                ]

        for fp in self.found:
            deserialzied = plistlib.load(fp(), fmt=plistlib.FMT_BINARY)

            scanned_networks = []
            try:
                for scanned_network in deserialzied[
                    "List of scanned networks with private mac"
                ]:
                    network = ScannedNetwork(scanned_network)
                    network.plist = fp().name
                    scanned_networks.append(network.attributes())
            except KeyError:
                self.log(logging.INFO, "-> No private networks scanned in plist file.")

        self.data = scanned_networks



import plistlib

from xleapp import Artifact, Search, WebIcon


class AppleMapsSearchHistory(Artifact):
    def __post_init__(self) -> None:
        self.name = "Apple Maps Search History"
        self.category = "Locations"
        self.web_icon = WebIcon.MAP_PIN
        self.report_headers = ("Tempstamp", "Search Entry")

    @Search(
        "*private/var/mobile/Containers/Data/Application/*/Library/Maps/GeoHistory.mapsdata"
    )
    def process(self) -> None:
        data_list = []
        for fp in self.found:
            plist_content = plistlib.load(fp())
            for entry in plist_content['MSPHistory']['records']:
                search_history = plist_content['MSPHistory']['records'][entry]
                content = search_history.get('contents')
                if content:
                    content = content.decode('UTF-8', 'ignore')
                else:
                    content = None
                timestamp = search_history.get('modificationDate')
                formatted_timestamp = timestamp.strftime('%Y-%m-%d %H:%M:%S')
                if content:
                    if len(content) < 300:
                        id_search_entry = content.split('\n')
                        search_entry = id_search_entry[1].split('"')
                        search_entry_split = str(search_entry[0]).split('\x12')
                        search_entry_filtered = list(filter(None, search_entry_split))
                        data_list.append(
                            (formatted_timestamp, ', '.join(search_entry_filtered))
                        )

        self.data = data_list



import datetime
import re


from xleapp import Artifact, Search, WebIcon


class AppConduit(Artifact):
    def __post_init__(self):

        self.name = "App Conduit"
        self.category = "App Conduit"
        self.web_icon = WebIcon.ACTIVITY
        self.report_headers = [
            ("Device ID", "Device type and version", "Device extra information"),
            ("Time", "Device interaction", "Device ID", "Log File Name"),
        ]

    @Search("**/AppConduit.log.*", file_names_only=True)
    def process(self):
        data_list = []
        device_type_and_info = []

        info = ""
        reg_filter = (
            r"(([A-Za-z]+[\s]+([a-zA-Z]+[\s]+[0-9]+)[\s]+"
            r"([0-9]+\:[0-9]+\:[0-9]+)[\s]+([0-9]{4}))([\s]+"
            r"[\[\d\]]+[\s]+[\<a-z\>]+[\s]+[\(\w\)]+)[\s\-]+"
            r"(((.*)(device+\:([\w]+\-[\w]+\-[\w]+\-[\w]+"
            r"\-[\w]+))(.*)$)))"
        )

        date_filter = re.compile(reg_filter)

        for fp in self.found:
            fp_found = open(fp(), "r", encoding="utf8")
            linecount = 0

            for line in fp_found:
                linecount = linecount + 1
                line_match = re.match(date_filter, line)

                if line_match:
                    date_time = line_match.group(3, 5, 4)
                    conv_time = " ".join(date_time)
                    dtime_obj = datetime.datetime.strptime(
                        conv_time,
                        "%b %d %Y %H:%M:%S",
                    )
                    values = line_match.group(9)
                    device_id = line_match.group(11)

                    if "devicesAreNowConnected" in values:
                        device = (
                            device_id,
                            line_match.group(12).split(" ")[4],
                            line_match.group(12).split(" ")[5],
                        )
                        device_type_and_info.append(device)

                        info = "Connected"
                        data = (dtime_obj, info, device_id, fp.name)
                        data_list.append(data)

                    if "devicesAreNoLongerConnected" in values:
                        info = "Disconnected"
                        data = (dtime_obj, info, device_id, fp.name)
                        data_list.append(data)
                    # if 'Resuming because' in values:
                    #     info = 'Resumed'
                    #     data_list.append((dtime_obj,info,device_id,device_type_tmp,file_name))
                    # if 'Suspending because' in values:
                    #     info = 'Suspended'
                    #     data_list.append((dtime_obj,info,device_id,device_type_tmp,file_name))
                    # if 'Starting reunion sync because device ' in values:
                    #     info = 'Reachable again after reunion sync'
                    #     data_list.append((dtime_obj,info,device_id,device_type_tmp,file_name))

        device_type_and_info = list(set(device_type_and_info))

        self.data = {
            "device_type_and_info": device_type_and_info,
            "device_connection_info": data_list,
        }



import re

from xleapp import Artifact, Search, WebIcon


class AppleWalletCards(Artifact):
    def __post_init__(self):
        self.name = "Cards"
        self.category = "Apple Wallet"
        self.web_icon = WebIcon.CREDIT_CARD
        self.report_headers = (
            'Timestamp (Card Added)',
            'Card Number',
            'Expiration Date',
            'Type',
        )

    @Search("*/com.apple.Passbook/Cache.db")
    def process(self) -> None:
        for fp in self.found:
            cursor = fp().cursor()
            cursor.execute(
                """
                SELECT
                TIME_STAMP,
                PROTO_PROPS
                FROM CFURL_CACHE_RESPONSE
                INNER JOIN CFURL_CACHE_BLOB_DATA ON CFURL_CACHE_BLOB_DATA.ENTRY_ID = CFURL_CACHE_RESPONSE.ENTRY_ID
                WHERE REQUEST_KEY LIKE '%CARDS'
                """,
            )

            all_rows = cursor.fetchall()

        if len(all_rows) > 0:
            data_list = []

            def get_bank_card_number(card_information):
                num_of_digits = [19, 18, 17, 16, 15, 14, 13]

                for digit_num in num_of_digits:
                    found_entry = re.findall(
                        r'\d{{{digits}}}'.format(digits=digit_num),
                        card_information,
                    )
                    if found_entry:
                        return found_entry[0]

            def get_card_type(card_num, num_length):
                first_digit = str(card_num)[:1]
                first_two_digits = str(card_num)[:2]

                if first_digit == '4' and (num_length == 13 or num_length == 16):
                    return 'Visa'
                elif first_digit == '5' and num_length == 16:
                    return 'Mastercard'
                elif first_digit == '6' and num_length == 16:
                    return 'Discover'
                elif (
                    first_two_digits == '34' or first_two_digits == '37'
                ) and num_length == 15:
                    return 'American Express'
                elif (
                    first_two_digits == '30'
                    or first_two_digits == '36'
                    or first_two_digits == '38'
                ) and num_length == 14:
                    return 'Diners Club Carte Blanche'
                else:
                    return 'Unknown'

            for row in all_rows:
                card_info = str(row[1], 'utf-8', 'ignore')
                card_number = get_bank_card_number(card_info)
                expiration_date = re.findall(r'\d{2}/\d{2}', card_info)
                card_type = get_card_type(card_number, len(card_number))

                data_list.append((row[0], card_number, expiration_date[0], card_type))

        self.data = data_list



import json

from pathlib import Path
from re import DOTALL, search

from xleapp import Artifact, Search, WebIcon


class AppleWalletPasses(Artifact):
    def __post_init__(self):
        self.name = "Passes"
        self.category = "Apple Wallet"
        self.web_icon = WebIcon.SEND
        self.report_headers = (
            'Unique ID',
            'Organization Name',
            'Type',
            'Localized Description',
            'Pass Added',
            'Pending Delete',
            'Pass Details',
            'Front Fields Content',
            'Back Fields Content',
            'Encoded Pass',
        )

    @Search("**/nanopasses.sqlite3", "**/Cards/*.pkpass/pass.json")
    def process(self) -> None:
        for fp in self.found:
            data_list = []
            json_files: dict[str, Path] = {}
            if fp.path.suffix == ".json":
                json_path: Path = fp.path
                unique_id = search(
                    r"(?:(?<=ards/)|(?<=ards\\))(.*?)(?=.pkpass)",
                    str(json_path.parent),
                    flags=DOTALL,
                ).group(0)
                filename = f"{str(unique_id)}_{str(json_path.name)}"
                self.copyfile(json_path, filename)

                json_files[filename] = json_path / filename

            if fp.path.suffix == ".sqlite3":
                cursor = fp().cursor()
                cursor.execute(
                    """
                        SELECT UNIQUE_ID, ORGANIZATION_NAME, TYPE_ID, LOCALIZED_DESCRIPTION,
                        DATETIME(INGESTED_DATE + 978307200,'UNIXEPOCH'), DELETE_PENDING, ENCODED_PASS,
                        FRONT_FIELD_BUCKETS, BACK_FIELD_BUCKETS
                        FROM PASS
                    """,
                )

                all_rows = cursor.fetchall()

                if len(all_rows) > 0:
                    for row in all_rows:
                        json_file = json_files.get(row[0], "")
                        if json_file:
                            try:
                                with open(json_file) as json_content:
                                    json_data = json.load(json_content)
                            except Exception:
                                json_data = "Malformed data"

                            encoded_pass = str(row[6], 'utf-8', 'ignore')
                            front_field = str(row[7], 'utf-8', 'ignore')
                            back_field = str(row[8], 'utf-8', 'ignore')
                            data_list.append(
                                (
                                    row[0],
                                    row[1],
                                    row[2],
                                    row[3],
                                    row[4],
                                    row[5],
                                    json_data,
                                    front_field,
                                    back_field,
                                    encoded_pass,
                                ),
                            )

        self.data = data_list



from xleapp import Artifact, WebIcon, Search


class AddressBook(Artifact):
    def __post_init__(self):
        self.name = "Address Book"
        self.category = "Address Book"
        self.web_icon = WebIcon.BOOK_OPEN
        self.report_headers = (
            "Contact ID",
            "Contact Number",
            "First Name",
            "Middle Name",
            "Last Name",
            "Creation Date",
            "Modification Date",
            "Storage Place",
        )

    @Search("**/AddressBook.sqlitedb")
    def process(self):
        for fp in self.found:
            cursor = fp().cursor()
            cursor.execute(
                """
                SELECT
                ABPerson.ROWID,
                VALUE,
                FIRST,
                MIDDLE,
                LAST,
                DATETIME(CREATIONDATE+978307200,'UNIXEPOCH'),
                DATETIME(MODIFICATIONDATE+978307200,'UNIXEPOCH'),
                NAME
                FROM ABPerson
                LEFT OUTER JOIN ABMultiValue ON ABPerson.ROWID = ABMultiValue.RECORD_ID
                LEFT OUTER JOIN ABStore ON ABPerson.STOREID = ABStore.ROWID
                """,
            )

            all_rows = cursor.fetchall()
            usageentries = len(all_rows)
            if usageentries > 0:
                data_list = []
                for row in all_rows:
                    an = str(row[0])
                    an = an.replace("b'", "")
                    an = an.replace("'", "")
                    data_list.append(
                        (
                            an,
                            row[1],
                            row[2],
                            row[3],
                            row[4],
                            row[5],
                            row[6],
                            row[7],
                        ),
                    )
                self.data = data_list



from xleapp import Artifact, Search, WebIcon


class ApplePodcastsShows(Artifact):
    def __post_init__(self):
        self.name = "Shows"
        self.category = "Apple Podcasts"
        self.web_icon = WebIcon.PLAY_CIRCLE
        self.report_headers = (
            'Date Added',
            'Date Last Played',
            'Date Last Updated',
            'Date Downloaded',
            'Author',
            'Title',
            'Feed URL',
            'Description',
            'Web Page URL',
        )

    @Search("**/MTLibrary.sqlite")
    def process(self) -> None:
        for fp in self.found:
            cursor = fp().cursor()
            cursor.execute(
                """
                select
                datetime(ZADDEDDATE + 978307200, 'unixepoch'),
                datetime(ZLASTDATEPLAYED + 978307200, 'unixepoch'),
                datetime(ZUPDATEDDATE + 978307200, 'unixepoch'),
                datetime(ZDOWNLOADEDDATE + 978307200, 'unixepoch'),
                ZAUTHOR,
                ZTITLE,
                ZFEEDURL,
                ZITEMDESCRIPTION,
                ZWEBPAGEURL
                from ZMTPODCAST
            """,
            )

            all_rows = cursor.fetchall()
        if len(all_rows) > 0:
            data_list = []
            for row in all_rows:
                data_list.append(
                    (
                        row[0],
                        row[1],
                        row[2],
                        row[3],
                        row[4],
                        row[5],
                        row[6],
                        row[7],
                        row[8],
                    ),
                )

        self.data = data_list



from xleapp import Artifact, Search, WebIcon


class GeodApplication(Artifact):
    def __post_init__(self):
        self.name = "GeoD Application"
        self.category = "Applications"
        self.web_icon = WebIcon.GRID
        self.report_headers = ("Creation Time", "Count ID", "Application")

    @Search("**/com.apple.geod/AP.db")
    def process(self):
        for fp in self.found:
            cursor = fp().cursor()
            cursor.execute(
                """
                SELECT count_type, app_id, createtime
                FROM mkcount
                """,
            )

            all_rows = cursor.fetchall()
            usageentries = len(all_rows)
            if usageentries > 0:
                data_list = []
                for row in all_rows:
                    data_list.append((row[2], row[0], row[1]))
                self.data = data_list



def hexify_byte(byte_to_convert):
    to_return = hex(byte_to_convert).replace("0x", "")
    if len(to_return) < 2:
        to_return += "0"

    return to_return


def bytes_to_mac_address(encoded_bytes):
    return (
        f"{hexify_byte(encoded_bytes[0])}:{hexify_byte(encoded_bytes[1])}:"
        f"{hexify_byte(encoded_bytes[2])}:{hexify_byte(encoded_bytes[3])}:"
        f"{hexify_byte(encoded_bytes[4])}:{hexify_byte(encoded_bytes[5])}"
    )



import base64
import gzip
import logging
import struct
import zlib

from xleapp import Artifact, Search, WebIcon


logger = logging.getLogger("xleapp.logfile")


class GeodMapTiles(Artifact):
    def __post_init__(self):
        self.name = "GeoD Map Tiles"
        self.category = "Geolocation"
        self.web_icon = WebIcon.MAP
        self.report_headers = (
            "Timestamp",
            "Places_from_VLOC",
            "Labels_in_tile",
            "Image",
            "Tileset",
            "Key A",
            "Key B",
            "Key C",
            "Key D",
            # "Size",
            # "ETAG"
        )

    @Search("**/com.apple.geod/MapTiles/MapTiles.sqlitedb")
    def process(self):
        for fp in self.found:
            cursor = fp().cursor()
            cursor.execute(
                """
                SELECT datetime(access_times.timestamp, 'unixepoch') as timestamp, key_a, key_b, key_c, key_d, tileset, data, size, etag
                FROM data
                INNER JOIN access_times on data.rowid = access_times.data_pk
                """,
            )

            all_rows = cursor.fetchall()
            usageentries = len(all_rows)
            data_list = []
            if usageentries > 0:
                for row in all_rows:
                    tcol_places = ""
                    vmp4_places = ""
                    data_parsed = ""

                    data = row["data"]
                    if data:  # NULL sometimes
                        if (
                            len(data) >= 11
                            and data[:11]
                            == b"\xff\xd8\xff\xe0\x00\x10\x4a\x46\x49\x46\x00"
                        ):
                            img_base64 = base64.b64encode(data).decode("utf-8")
                            img_html = f'<img src="data:image/jpeg;base64, {img_base64}" alt="Map Tile" />'
                            data_parsed = img_html
                        elif len(data) >= 4 and data[:4] == b"TCOL":
                            vmp4_places, tcol_places = parsetcol(data)
                            vmp4_places = ", ".join(vmp4_places)
                            tcol_places = ", ".join(tcol_places)
                        elif len(data) >= 4 and data[:4] == b"VMP4":
                            vmp4_places = parsevmp4(data)
                            vmp4_places = ", ".join(vmp4_places)
                    # else:
                    # header_bytes = data[:28]
                    # hexdump = generate_hexdump(header_bytes, 5) if header_bytes else ''
                    # data_parsed = hexdump

                    data_list.append(
                        (
                            row["timestamp"],
                            tcol_places,
                            vmp4_places,
                            data_parsed,
                            get_hex(row["tileset"]),
                            get_hex(row["key_a"]),
                            get_hex(row["key_b"]),
                            get_hex(row["key_c"]),
                            get_hex(row["key_d"]),
                        ),
                    )
                self.data = data_list


def readvloc(data) -> list:
    names = []
    total_len = len(data)
    pos = 8
    while pos < total_len:
        if data[pos] < 0x80:
            skip_len = 2
        else:
            skip_len = 3
        end_pos = data[pos + skip_len :].find(b"\0")
        if end_pos >= 0:
            name = data[pos + skip_len : pos + skip_len + end_pos].decode(
                "utf8",
                "ignore",
            )
            if name:
                names.append(name)
            pos += skip_len + end_pos + 1
        else:
            break
    return names


def parsetcol(data) -> tuple:
    tcol_places = []
    data_size = len(data)
    if data_size >= 8:
        tcol_data_offset = struct.unpack("<I", data[4:8])[0]
        tcol_compressed_data = data[tcol_data_offset:]
        if tcol_compressed_data:
            try:
                tcol_places = gzip.decompress(tcol_compressed_data)
            except (OSError, EOFError, zlib.error) as ex:
                logger.info(f"Gzip decompression error from parsetcol() - {str(ex)}")
                tcol_places = ""
        vmp4_places = parsevmp4(data[8:tcol_data_offset])
        return vmp4_places, readvloc(tcol_places)


def parsevmp4(data):
    num_items = struct.unpack("<H", data[6:8])[0]
    pos = 8
    for _item in range(num_items):
        item_type, offset, size = struct.unpack("<HII", data[pos : pos + 10])
        if item_type == 10:
            item_data = data[offset : offset + size]
            if item_data[0] == 1:
                compressed_data = item_data[5:]
                try:
                    places_data = zlib.decompress(compressed_data)
                except zlib.error as ex:
                    logging.error(
                        f"Zlib decompression error from parsevmp4() - {str(ex)}",
                    )
                    places_data = ""
            else:
                places_data = item_data[1:]
            return [
                x.decode("UTF8", "ignore") for x in places_data.rstrip(b"\0").split(b"\0")
            ]
        pos += 10
    return []


def get_hex(num):
    if num:
        return hex(num).upper()
    return ""



import pathlib
import plistlib
import sys


import biplist
from xleapp import Artifact, WebIcon, Search


class AppGroupListing(Artifact):
    def __post_init__(self):
        self.name = "App Group Listing"
        self.description = (
            "List can included once installed but not present "
            "apps. Each file is named .com.apple.mobile_"
            "container_manager.metadata.plist"
        )
        self.category = "Installed Apps"
        self.report_headers = ("Bundle ID", "Type", "Directory GUID", "Path")
        self.report_title = "Bundle ID by AppGroup & PluginKit IDs"
        self.web_icon = WebIcon.PACKAGE

    @Search(
        "*/private/var/mobile/Containers/Shared/AppGroup/*/*.metadata.plist",
        "**/PluginKitPlugin/*.metadata.plist",
    )
    def process(self):
        data_list = []

        for fp in self.found:
            plist = plistlib.load(fp())
            bundleid = plist["MCMMetadataIdentifier"]

            path = fp.path
            appgroupid = path.parent.name
            fileloc = path.parent
            typedir = path.parents[1].name

            data_list.append((bundleid, typedir, appgroupid, fileloc))

        self.data = data_list



import plistlib

from datetime import datetime

from xleapp import Artifact, Search, WebIcon


class VoiceRecording(Artifact):
    def __post_init__(self):
        self.name = "Voice Recordings"
        self.category = "Voice-Recordings"
        self.report_headers = ("Creation Date", "Title", "Path to File", "Audio File")
        self.report_title = "Voice Recordings"
        self.web_icon = WebIcon.MIC

    @Search(
        "**/Recordings/*.composition/manifest.plist",
        "**/Recordings/*.m4a",
        file_names_only=True,
    )
    def process(self):
        data_list = []
        plist_files = []
        m4a_files = []

        for file_found in self.found:
            if file_found().suffix(".plist"):
                plist_files.append(file_found())
            elif file_found.suffix(".m4a"):
                m4a_filename = file_found()
                if " " in m4a_filename:
                    m4a_filename = m4a_filename.replace(" ", "_")
                m4a_files.append(m4a_files)
                self.copyfile(file_found, m4a_filename)

        plist_files.sort()
        m4a_files.sort()

        for plist_file, m4a_file in zip(plist_files, m4a_files):
            with open(plist_file, "rb") as file:
                pl = plistlib.load(file)
                ct = unix_epoch_to_readable_date(pl["RCSavedRecordingCreationTime"])

                audio = (
                    "<audio controls>"
                    f'<source src="{m4a_file}" type="audio/wav">'
                    "<p>Your browser does not support HTML5"
                    "audio elements.</p></audio>"
                )

                data_list.append(
                    (
                        ct,
                        pl["RCSavedRecordingTitle"],
                        pl["RCComposedAVURL"].split("//")[1],
                        audio,
                    ),
                )

        self.data = data_list


def unix_epoch_to_readable_date(unix_epoch_time):
    unix_time = float(unix_epoch_time + 978307200)
    readable_time = datetime.utcfromtimestamp(unix_time).strftime("%Y-%m-%d %H:%M:%S")
    return readable_time



import logging
import plistlib

from xleapp import Artifact, Search, WebIcon


class AppleWifiKnownNetworks(Artifact):
    def __post_init__(self):
        self.name = "Wifi Known Networks"
        self.description = (
            "WiFi known networks data. Dates are taken straight from the source plist."
        )
        self.category = "Locations"
        self.report_headers = (
            "SSID",
            "BSSID",
            "Network Usage",
            "Country Code",
            "Device Name",
            "Manufacturer",
            "Serial Number",
            "Model Name",
            "Last Joined",
            "Last Auto Joined",
            "Last Updated",
            "Enabled",
            "WiFi Network Password Modification Date",
            "File",
        )
        self.report_title = "WiFi Known Networks"
        self.web_icon = WebIcon.WIFI

    @Search(
        "**/com.apple.wifi.plist",
        "**/com.apple.wifi-networks.plist.backup",
        "**/com.apple.wifi.known-networks.plist",
        "**/com.apple.wifi-private-mac-networks.plist",
    )
    def process(self):
        class KnownNetwork:
            ssid: str
            bssid: str
            net_usage: str
            country_code: str
            device_name: str = ""
            manufacturer: str = ""
            serial_number: str = ""
            model_name: str = ""
            last_joined: str
            last_updated: str
            last_auto_joined: str
            enabled: str
            wnpmd: str
            plist: str

            def __init__(self, network: dict) -> None:

                self.ssid = str(network.get("SSID_STR", ""))
                self.bssid = str(network.get("BSSID", ""))
                self.net_usage = str(network.get("networkUsage", ""))
                self.country_code = str(
                    network.get("80211D_IE", {}).get("IE_KEY_80211D_COUNTRY_CODE", ""),
                )
                self.last_updated = str(network.get("lastUpdated", ""))
                self.last_joined = str(network.get("lastJoined", ""))
                self.last_auto_joined = str(network.get("lastAutoJoined", ""))
                self.wnpmd = str(
                    network.get("WiFiNetworkPasswordModificationDate", ""),
                )
                self.enabled = network.get("enabled", "")

                wps_prob_resp_ie = network.get("WPS_PROB_RESP_IE", "")
                if wps_prob_resp_ie:
                    self.device_name = wps_prob_resp_ie.get("IE_KEY_WPS_DEV_NAME", "")
                    self.manufacturer = wps_prob_resp_ie.get(
                        "IE_KEY_WPS_MANUFACTURER", ""
                    )
                    self.serial_number = wps_prob_resp_ie.get("IE_KEY_WPS_SERIAL_NUM", "")
                    self.model_name = wps_prob_resp_ie.get(
                        "IE_KEY_WPS_MODEL_NAME",
                        "",
                    )

            def attributes(self) -> list[str]:
                return [
                    self.ssid,
                    self.bssid,
                    self.net_usage,
                    self.country_code,
                    self.device_name,
                    self.manufacturer,
                    self.serial_number,
                    self.model_name,
                    self.last_joined,
                    self.last_updated,
                    self.last_auto_joined,
                    self.enabled,
                    self.wnpmd,
                    self.plist,
                ]

        for fp in self.found:
            deserialized = plistlib.load(fp())

            known_networks = []
            try:
                for known_network in deserialized["List of known networks"]:
                    network = KnownNetwork(known_network)
                    network.plist = fp().name
                    known_networks.append(network.attributes())
            except KeyError:
                self.log(logging.INFO, "-> No networks found in plist.")

        self.data = known_networks



import io
import plistlib

import nska_deserialize as nd

from xleapp import Artifact, Search, WebIcon


class ApplicationState(Artifact):
    def __post_init__(self):
        self.name = "Application State"
        self.category = "Installed Apps"
        self.web_icon = WebIcon.PACKAGE
        self.report_headers = ("Bundle ID", "Bundle Path", "Sandbox Path")

    @Search("**/applicationState.db")
    def process(self) -> None:
        for fp in self.found:
            cursor = fp().cursor()
            cursor.execute(
                """
                select ait.application_identifier as ai, kvs.value as compat_info,
                (SELECT kvs.value from kvs left join application_identifier_tab on application_identifier_tab.id = kvs.application_identifier
                left join key_tab on kvs.key = key_tab.id
                WHERE key_tab.key="XBApplicationSnapshotManifest" and kvs.key = key_tab.id
                and application_identifier_tab.id = ait.id
                ) as snap_info
                from kvs
                left join application_identifier_tab ait on ait.id = kvs.application_identifier
                left join key_tab on kvs.key = key_tab.id
                where key_tab.key="compatibilityInfo"
                order by ait.id
                """,
            )

        all_rows = cursor.fetchall()
        if len(all_rows) > 0:
            data_list = []
            snap_info_list = []
            for row in all_rows:
                plist_file_object = io.BytesIO(row[1])
                if row[1].find(b"NSKeyedArchiver") == -1:
                    plist = plistlib.load(plist_file_object)
                else:
                    try:
                        plist = nd.deserialize_plist(plist_file_object)
                    except (
                        nd.DeserializeError,
                        nd.biplist.NotBinaryPlistException,
                        nd.biplist.InvalidPlistException,
                        nd.plistlib.InvalidFileException,
                        nd.ccl_bplist.BplistError,
                        ValueError,
                        TypeError,
                        OSError,
                        OverflowError,
                    ) as ex:
                        self.log(
                            message=(
                                f"Failed to read plist for {row[0]}, "
                                f"error was: {ex!r}"
                            ),
                        )
                if plist:
                    if type(plist) is dict:
                        var1 = plist.get("bundleIdentifier", "")
                        var2 = plist.get("bundlePath", "")
                        var3 = plist.get("sandboxPath", "")
                        data_list.append((var1, var2, var3))
                        if row[2]:
                            snap_info_list.append((var1, var2, var3, row[2]))
                    else:
                        self.log(
                            message=(
                                f"For {row[0]} Unexpected type {type(plist)!r} "
                                "found as plist root, can't process"
                            ),
                        )
                else:
                    self.log(message=f"For {row[0]!r}, plist could not be read!")



import plistlib

from xleapp import Artifact, Search, WebIcon


class Alarms(Artifact):
    def __post_init__(self) -> None:
        self.name = "Alarms"
        self.category = "Alarms"
        self.web_icon = WebIcon.CLOCK
        self.report_headers = (
            'Alarm Title',
            'Alarm Enabled',
            'Fire Date',
            'Dismiss Date',
            'Last Modified',
            'Repeat Schedule',
            'Alarm Sound',
            'Is Sleep Alarm',
            'Bedtime Not Disturbed',
            'Bedtime Fire Date',
        )

    @Search("*private/var/mobile/Library/Preferences/com.apple.mobiletimerd.plist")
    def process(self) -> None:
        for fp in self.found:
            pl = plistlib.load(fp())

            data_list = []
            if 'MTAlarms' in pl:
                if 'MTAlarms' in pl['MTAlarms']:
                    for alarms in pl['MTAlarms']['MTAlarms']:
                        alarms_dict: dict = alarms['$MTAlarm']

                        alarm_title = alarms_dict.get('MTAlarmTitle', 'Alarm')
                        fire_date = alarms_dict.get('MTAlarmFireDate', '')
                        dismiss_date = alarms_dict.get('MTAlarmDismissDate', '')
                        repeat_schedule = decode_repeat_schedule(
                            alarms_dict['MTAlarmRepeatSchedule']
                        )

                        data_list.append(
                            (
                                alarm_title,
                                alarms_dict['MTAlarmEnabled'],
                                fire_date,
                                dismiss_date,
                                alarms_dict['MTAlarmLastModifiedDate'],
                                ', '.join(repeat_schedule),
                                alarms_dict['MTAlarmSound']['$MTSound']['MTSoundToneID'],
                                alarms_dict['MTAlarmIsSleep'],
                                alarms_dict['MTAlarmBedtimeDoNotDisturb'],
                                '',
                            )
                        )

                if 'MTSleepAlarm' in pl['MTAlarms']:
                    for sleep_alarms in pl['MTAlarms']['MTSleepAlarm']:
                        sleep_alarm_dict: dict = pl['MTAlarms']['MTSleepAlarm'][
                            sleep_alarms
                        ]

                        alarm_title = sleep_alarm_dict.get('MTAlarmTitle', 'Bedtime')

                        repeat_schedule = decode_repeat_schedule(
                            sleep_alarm_dict['MTAlarmRepeatSchedule']
                        )

                        data_list.append(
                            (
                                alarm_title,
                                sleep_alarm_dict['MTAlarmEnabled'],
                                sleep_alarm_dict['MTAlarmFireDate'],
                                sleep_alarm_dict['MTAlarmDismissDate'],
                                sleep_alarm_dict['MTAlarmLastModifiedDate'],
                                ', '.join(repeat_schedule),
                                sleep_alarm_dict['MTAlarmSound']['$MTSound'][
                                    'MTSoundToneID'
                                ],
                                sleep_alarm_dict['MTAlarmIsSleep'],
                                sleep_alarm_dict['MTAlarmBedtimeDoNotDisturb'],
                                sleep_alarm_dict['MTAlarmBedtimeFireDate'],
                            )
                        )

        self.data = data_list


def decode_repeat_schedule(repeat_schedule_value):
    days_list = {
        64: 'Sunday',
        32: 'Saturday',
        16: 'Friday',
        8: 'Thursday',
        4: 'Wednesday',
        2: 'Tuesday',
        1: 'Monday',
    }
    schedule = []

    if repeat_schedule_value == 127:
        schedule.append('Every Day')
        return schedule
    elif repeat_schedule_value == 0:
        schedule.append('Never')
        return schedule

    for day in days_list:
        if repeat_schedule_value > 0 and repeat_schedule_value >= day:
            repeat_schedule_value -= day
            schedule.append(days_list[day])
    return reversed(schedule)



import plistlib

import blackboxprotobuf

from xleapp import Artifact, Search, WebIcon


class AppleMapsGroup(Artifact):
    def __post_init__(self) -> None:
        self.name = "Apple Maps Group"
        self.category = "Locations"
        self.web_icon = WebIcon.MAP_PIN
        self.report_headers = ('Latitude', 'Longitude')

    @Search("**/Shared/AppGroup/*/Library/Preferences/group.com.apple.Maps.plist")
    def process(self) -> None:
        for fp in self.found:
            deserialized_plist = plistlib.load(fp())
            types = {
                '1': {
                    'type': 'message',
                    'message_typedef': {
                        '1': {'type': 'int', 'name': ''},
                        '2': {'type': 'int', 'name': ''},
                        '5': {
                            'type': 'message',
                            'message_typedef': {
                                '1': {'type': 'double', 'name': 'Latitude'},
                                '2': {'type': 'double', 'name': 'Longitude'},
                                '3': {'type': 'double', 'name': ''},
                                '4': {'type': 'fixed64', 'name': ''},
                                '5': {'type': 'double', 'name': ''},
                            },
                            'name': '',
                        },
                        '7': {'type': 'int', 'name': ''},
                    },
                    'name': '',
                },
            }
            try:
                internal_deserialized_plist, di = blackboxprotobuf.decode_message(
                    (deserialized_plist['MapsActivity']), types
                )
            except KeyError:
                self.log(message="-> No data in Apple Maps Groups!")
                return

            latitude = internal_deserialized_plist['1']['5']['Latitude']
            longitude = internal_deserialized_plist['1']['5']['Longitude']

            data_list = []
            data_list.append((latitude, longitude))
        self.data = data_list



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['xleapp_ios',
 'xleapp_ios.helpers',
 'xleapp_ios.helpers.parsers',
 'xleapp_ios.helpers.parsers.ccl',
 'xleapp_ios.helpers.parsers.ktx',
 'xleapp_ios.plugins']

package_data = \
{'': ['*']}

install_requires = \
['astc-decomp>=1.0.3,<2.0.0',
 'astc_decomp>=1.0.3,<2.0.0',
 'blackboxprotobuf>=1.0.1,<2.0.0',
 'bplist>=1.1,<2.0',
 'bs4>=0.0.1,<0.0.2',
 'nska-deserialize>=1.3.1',
 'pandas>=1.3.3,<2.0.0',
 'pillow>=8.3.2,<9.0.0',
 'pyliblzfse>=0.4.1,<0.5.0']

entry_points = \
{'xleapp.plugins': ['ios = xleapp_ios.plugins:IosPlugin']}

setup_kwargs = {
    'name': 'xleapp-ios',
    'version': '0.1.1',
    'description': 'iOS Artifacts for xLEAPP (free)',
    'long_description': '# xLEAPP iOS Artifacts \n\nProvides iOS artifacts supported under [MIT license ](https://opensource.org/licenses/MIT).\n\nThis is not a standalone package. xLEAPP is required to be installed. See xLEAPP project documentation for information on using this plugins.',
    'author': 'Jesse Spangenberger',
    'author_email': 'azuleonyx@digitalforensics.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/flamusdiu/xleapp',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import fnmatch
import logging
import shutil
import sqlite3
import tarfile
import typing as t
from functools import cached_property
from pathlib import Path
from zipfile import ZipFile

from xleapp.artifacts import ArtifactError, Artifacts
from xleapp.helpers.db import open_sqlite_db_readonly
from xleapp.helpers.search import FileSearchProvider, FileSeekerBase
from xleapp.helpers.utils import is_platform_windows
from xleapp.plugins import Plugin

logger = logging.getLogger("xleapp.logfile")


class IosPlugin(Plugin):
    def __init__(self) -> None:
        super().__init__()

    @property
    def folder(self) -> Path:
        """Returns path of the plugin folder"""
        return Path(__file__).parent

    def pre_process(self, artifacts: Artifacts) -> None:
        for artifact in artifacts.data:
            # Now ready to run
            # Special processing for iTunesBackup Info.plist as it is a
            # separate entity, not part of the Manifest.db. Seeker won't find it
            if (
                artifacts.app.device["type"] == "ios"
                and artifacts.app.extraction_type == "ITUNES"
            ):
                if artifact.name == "ITUNES_BACKUP_INFO":
                    info_plist_path = Path(artifacts.app.input_path) / "Info.plist"
                    if not info_plist_path.exists():
                        ArtifactError("Info.plist not found for iTunes Backup!")
                    else:
                        artifacts['LAST_BUILD'].select = False
            else:
                artifacts['ITUNES_BACKUP_INFO'].select = False

    def register_seekers(self, search_providers: FileSearchProvider) -> None:
        search_providers.register_builder("ITUNES", FileSeekerItunes())


class FileSeekerItunes(FileSeekerBase):
    """Searches iTunes Backup for files."""

    manifest_db: Path
    directory: Path

    def __call__(
        self,
        directory_or_file,
        temp_folder,
    ) -> t.Type[FileSeekerBase]:
        self.input_path = Path(directory_or_file)
        self.temp_folder = Path(temp_folder)
        if self.validate:
            self.all_files = self.build_files_list()

        return self

    def build_files_list(self, folder=None) -> list:
        """Populates paths from Manifest.db files into _all_files"""

        all_files: dict[str, str] = {}
        db = open_sqlite_db_readonly(self.manifest_db)
        cursor = db.cursor()
        cursor.execute(
            """
            SELECT
            fileID,
            relativePath
            FROM
            Files
            WHERE
            flags=1
            """,
        )
        db.row_factory = sqlite3.Row
        all_rows = cursor.fetchall()
        for row in all_rows:
            relative_path: str = row['relativePath']
            hash_filename: str = row['fileID']
            all_files[relative_path] = hash_filename
        db.close()

        return all_files

    def search(self, filepattern: str):
        pathlist = []

        if filepattern.find("*") == -1:
            if is_platform_windows():
                original_location = Path(f"\\\\?\\{self.directory / filepattern}")
            else:
                original_location = self.directory / filepattern
            pathlist.append(original_location)
        else:
            matching_keys = fnmatch.filter(self._all_files, filepattern)
            for relative_path in matching_keys:
                hash_filename = self._all_files[relative_path]

                if is_platform_windows():
                    original_location = Path(
                        f"\\\\?\\{self.directory / hash_filename[:2] / hash_filename}"
                    )
                    temp_location = Path(f"\\\\?\\{self.temp_folder / relative_path}")
                else:
                    original_location = (
                        Path(self.directory) / hash_filename[:2] / hash_filename
                    )
                    temp_location = Path(self.temp_folder) / relative_path
                temp_location.parent.mkdir(parents=True, exist_ok=True)
                shutil.copyfile(original_location, temp_location)
                pathlist.append(temp_location)

        return iter(pathlist)

    def cleanup(self) -> None:
        # no cleanup for this seeker
        pass

    @cached_property
    def validate(self) -> bool:
        mime: str
        input_path: Path
        mime, input_path = self.input_path
        extract_dir = self.temp_folder / "backup"

        if mime == "dir":
            manifest_db = input_path / "Manifest.db"
            if manifest_db.exists():
                self.manifest_db = manifest_db
        elif mime in ["application/x-gzip", "application/x-tar"]:
            with tarfile.open(input_path, "r:*") as gz_or_tar_file:
                num_of_files = sum(1 for member in gz_or_tar_file if member.isreg())
                for member in gz_or_tar_file:
                    if fnmatch.fnmatch(member.name, "**/Manifest.db"):
                        logger.info(f"Manifest.db found in {input_path!r}...")
                        logger.info(
                            f"Extracing {num_of_files} files from backup to {self.temp_folder!r}"
                        )

                        gz_or_tar_file.extractall(path=extract_dir)
                        self.manifest_db = extract_dir / member.path
        elif mime == "application/zip":
            with ZipFile(input_path) as zip_file:
                num_of_files = len(zip_file.infolist())
                for member in zip_file.namelist():
                    if fnmatch.fnmatch(member, "**/Manifest.db"):
                        logger.info(f"Manifest.db found in {input_path!r}...")
                        logger.info(
                            f"Extracing {num_of_files} files from backup to {self.temp_folder!r}"
                        )
                        zip_file.extractall(path=extract_dir)
                        self.manifest_db = extract_dir / member

        if self.manifest_db:
            self.directory = self.manifest_db.parent
            return True
        return False

    @property
    def priorty(self) -> int:
        return 10



from collections import defaultdict

from pathlib import Path

from xleapp import Artifact, Search, WebIcon, open_sqlite_db_readonly


class WhatsappMessages(Artifact):
    def __post_init__(self):
        self.name = "Whatsapp - Messages"
        self.category = "Whatsapp"
        self.report_headers = (
            "Timestamp",
            "Sender Name",
            "From ID",
            "Receiver",
            "To ID",
            "Message",
            "Attachment File",
            "Thumb",
            "Attachment Local Path",
            "Starred?",
            "Latitude",
            "Longitude",
        )
        self.web_icon = WebIcon.MESSAGE_SQUARE
        self.kml = True
        self.timeline = True

    @Search(
        "*/var/mobile/Containers/Shared/AppGroup/*/ChatStorage.sqlite*",
        "*/var/mobile/Containers/Shared/AppGroup/*/Message/Media/*/*/*/*.*",
        file_names_only=True,
        return_on_first_hit=False,
    )
    def process(self):
        data_list = []
        media_list = defaultdict(set)
        files_found = self.found.copy()
        for file_found in files_found:
            if not file_found().suffix.startswith(".sqlite"):
                media_list[file_found().stem].add(file_found())
                self.found.remove(file_found)

            if file_found().name.endswith(".sqlite"):
                fp = str(file_found())

        db = open_sqlite_db_readonly(fp)
        cursor = db.cursor()

        cursor.execute(
            """
            select
            datetime(ZMESSAGEDATE+978307200, 'UNIXEPOCH'),
            ZISFROMME,
            ZPARTNERNAME,
            ZFROMJID,
            ZTOJID,
            ZWAMESSAGE.ZMEDIAITEM,
            ZTEXT,
            ZSTARRED,
            ZMESSAGETYPE,
            ZLONGITUDE,
            ZLATITUDE,
            ZMEDIALOCALPATH,
            ZXMPPTHUMBPATH
            FROM ZWAMESSAGE
            left JOIN ZWAMEDIAITEM
            on ZWAMESSAGE.Z_PK = ZWAMEDIAITEM.ZMESSAGE 
            left JOIN ZWACHATSESSION
            on ZWACHATSESSION.Z_PK = ZWAMESSAGE.ZCHATSESSION
            """
        )
        all_rows = cursor.fetchall()
        usageentries = len(all_rows)

        if usageentries > 0:
            for row in all_rows:

                if row[1] == 1:
                    sender = "Local User"
                    receiver = row[2]
                else:
                    sender = row[2]
                    receiver = "Local User"

                if row[8] == 5:
                    lon = row[9]
                    lat = row[10]
                else:
                    lat = ""
                    lon = ""

                thumb = row[12]
                attachment = row[11]
                localpath = row[11]

                if thumb and Path(thumb).stem in media_list:
                    media_name = Path(thumb).name
                    media_base_name = Path(thumb).stem
                    media_base_path = media_list[media_base_name].pop().parent

                    try:
                        thumb_path = media_base_path / f"{media_name}"
                        url_src_path = self.copyfile(thumb_path, media_name)
                        thumb = f'<img src="{url_src_path}"></img>'
                    except FileNotFoundError:
                        thumb = ""
                else:
                    thumb = ""

                if attachment and Path(attachment).stem in media_list:
                    media_name = Path(attachment).name
                    media_base_name = Path(attachment).stem
                    media_base_path = media_list[media_base_name].pop().parent

                    try:
                        attachment_path = media_base_path / f"{media_name}"
                        url_src_path = self.copyfile(attachment_path, media_name)
                        attachment = f'<img src="{url_src_path}" width="300"></img>'
                    except FileNotFoundError:
                        attachment = ""
                else:
                    attachment = ""

                data_list.append(
                    (
                        row[0],
                        sender,
                        row[3],
                        receiver,
                        row[4],
                        row[6],
                        attachment,
                        thumb,
                        localpath,
                        row[7],
                        lat,
                        lon,
                    )
                )

        self.data = data_list



from xleapp import Artifact, Search, WebIcon


class Accounts(Artifact):
    def __post_init__(self):
        self.name = "Accounts"
        self.category = "Accounts"
        self.web_icon = WebIcon.USER
        self.report_headers = (
            "Timestamp",
            "Account Desc.",
            "Username",
            "Description",
            "Identifier",
            "Bundle ID",
        )

    @Search("**/Accounts3.sqlite")
    def process(self):
        for fp in self.found:
            cursor = fp().cursor()
            cursor.execute(
                """
                select
                datetime(zdate+978307200,'unixepoch','utc' ),
                zaccounttypedescription,
                zusername,
                zaccountdescription,
                zaccount.zidentifier,
                zaccount.zowningbundleid
                from zaccount, zaccounttype
                where zaccounttype.z_pk=zaccount.zaccounttype
                """,
            )

            all_rows = cursor.fetchall()
            usageentries = len(all_rows)
            if usageentries > 0:
                data_list = []
                for row in all_rows:
                    data_list.append((row[0], row[1], row[2], row[3], row[4], row[5]))
                self.data = data_list
